@@ -20,7 +20,9 @@
 
 #include <kiconloader.h>
 
+#include <KIconTheme>
 #include <QPainter>
+#include <QSet>
 
 KIconEngine::KIconEngine(const QString &iconName, KIconLoader *iconLoader, const QStringList &overlays)
     : mIconName(iconName),
@@ -49,6 +51,9 @@ static inline int qIconModeToKIconState(QIcon::Mode mode)
     case QIcon::Disabled:
         kstate = KIconLoader::DisabledState;
         break;
+    case QIcon::Selected:
+        kstate = KIconLoader::SelectedState;
+        break;
     }
     return kstate;
 }
@@ -69,8 +74,10 @@ void KIconEngine::paint(QPainter *painter, const QRect &rect, QIcon::Mode mode,
 
     Q_UNUSED(state)
 
+    qreal dpr = painter->device()->devicePixelRatio();
+
     const int kstate = qIconModeToKIconState(mode);
-    const int iconSize = qMin(rect.width(), rect.height());
+    const int iconSize = qMin(rect.width(), rect.height()) * dpr;
     const QPixmap pix = mIconLoader.data()->loadIcon(mIconName, KIconLoader::Desktop, iconSize, kstate, mOverlays);
     painter->drawPixmap(rect, pix);
 }
@@ -79,12 +86,17 @@ QPixmap KIconEngine::pixmap(const QSize &size, QIcon::Mode mode, QIcon::State st
 {
     Q_UNUSED(state)
 
+    if (!size.isValid()) {
+        return QPixmap();
+    }
+
     if (!mIconLoader) {
         QPixmap pm(size);
         pm.fill(Qt::transparent);
         return pm;
     }
 
+
     const int kstate = qIconModeToKIconState(mode);
     const int iconSize = qMin(size.width(), size.height());
     QPixmap pix = mIconLoader.data()->loadIcon(mIconName, KIconLoader::Desktop, iconSize, kstate, mOverlays);
@@ -107,6 +119,8 @@ QString KIconEngine::iconName() const
     return mIconName;
 }
 
+Q_GLOBAL_STATIC_WITH_ARGS(QList<QSize>, sSizes, (QList<QSize>() << QSize(16, 16) << QSize(22, 22) << QSize(32, 32) << QSize(48, 48) << QSize(64, 64) << QSize(128, 128) << QSize(256, 256)))
+
 QList<QSize> KIconEngine::availableSizes(QIcon::Mode mode, QIcon::State state) const
 {
     Q_UNUSED(mode);
@@ -116,22 +130,13 @@ QList<QSize> KIconEngine::availableSizes(QIcon::Mode mode, QIcon::State state) c
         return QList<QSize>();
     }
 
-    if (mIconLoader->iconPath(mIconName, KIconLoader::Desktop, KIconLoader::MatchBest).isEmpty()) {
-        return QList<QSize>();
-    }
-
-    return QList<QSize>() << QSize(16, 16)
-           << QSize(22, 22)
-           << QSize(32, 32)
-           << QSize(48, 48)
-           << QSize(64, 64)
-           << QSize(128, 128)
-           << QSize(256, 256);
+    bool found = mIconLoader->hasIcon(iconName());
+    return found ? *sSizes : QList<QSize>();
 }
 
 QString KIconEngine::key() const
 {
-    return QString::fromLatin1("KIconEngine");
+    return QStringLiteral("KIconEngine");
 }
 
 QIconEngine *KIconEngine::clone() const
